/*
 * VAX-11 instruction disassembler
 * (c) by D. Pryanishnikov 5/7/1996
 *
 *  3-Dec-1998 Some corrections for GCC
 *
 * void disvini(void) - initialization. Should be called whenever
 *                      current location pointer changes voluntary
 *                      (e.g. not equal to result of previous call of
 *                      disassembler)
 *
 * long disvax(outbuf,getfwd,point) - disassemble an instruction
 *
 * Inputs:
 * char *outbuf - output (ASCII) buffer pointer
 * char (*getfwd)() - user-defined function to get next byte
 *                    called with the argument of current virtual address
 * long point - virtual address of instruction
 *
 * Output line in outbuf will be zero-terminated
 * Function returns next virtual address
 *
 * Notes:
 * o CASEx instruction may cause mistrack in case of non-literal and
 *   non-immediate third operand. No additional fetch will be attempted
 *   to fetch LIMIT value (since it is incorrect attempt).
 * o Floating-point operands are represented in integer notation
 */

#define NULL (char *)0

extern char *vaxinst[256];
extern char *vaxFD[256];

long localpoint;
unsigned char (*localgetfwd)();
char *localoutbuf;

static isincase=0;      /* Flag: non-zero means "doing CASEx offsets" */

static isindex=0;       /* Flag: nonzero means "parsing indexed method" */
static leading=0;       /* Flag: non-zero means "don't supress leading zeros" */

static isvalid=0;       /* Flag: nonzero means "last arg. was immediate" */

static long lastval;    /* In case of isvalid!=0 value of last argument */
static long svpoint;    /* -->>-- saved address of the first offset */

disvini()
{ isincase = 0; }

static unsigned char getnext()
{ return (*localgetfwd)(localpoint++); }

split32(input,quo,rem)
long input;
long *quo,*rem;
{
 *quo = (input>>4) & 0x0FFFFFFF;
 *rem = input &0xF; 
}

static outchar(chr)
char chr;
{ *localoutbuf++ = chr; }

static outstr(str)
register char *str;
{ while (*str) *localoutbuf++ = *str++ ; }

static outnum(lnum,digcnt)
long lnum;
int digcnt;
{ 
 long tmprem,tmpquo;
 split32(lnum,&tmpquo,&tmprem);
 if ( (tmpquo || leading) && --digcnt) outnum(tmpquo,digcnt);
 outchar( (char)(tmprem>9 ? tmprem+'a'-10 : tmprem+'0'));
} 

/*
 * Immediate operand retrieve functions
 */

long getib()
{ char i = getnext(); 
  return i; }

long getiw()
{ short i = getnext();
        i |= (short)getnext()<<8;
  return i; }

long getil()
{   long tmp=getnext();
         tmp += ((long)getnext()<<8);
         tmp += ((long)getnext()<<16);
  return tmp +  ((long)getnext()<<24); }

/*
 * The main function 
 */

long disvax(outbuf,getfwd,point)
char *outbuf;
unsigned char (*getfwd)();
long point;
{
 register char *cptr;
 register unsigned curbyt;
 int i;

 localgetfwd=getfwd; localpoint=point; localoutbuf=outbuf; 

 isindex = 0;           /* Not in indexed expression */
 leading = 0;           /* Supress leading zeros */

 if (isincase)          /* Pending CASEx context */
  {
   outstr(".word   ");     
   outnum(svpoint+getiw(),8);     
   outchar('-');
   outnum(svpoint);
   if ( !(lastval--) ) isincase = 0;
   goto end;
  }

 if ( (curbyt=getnext()) == 0xFD )
  {
   if ( !(cptr=vaxFD[curbyt=getnext()]) )
     { outstr(".byte   "); outnum((long)0xFD,2);
       outchar(','); outnum((long)curbyt,2); goto end; }
  }    
 else if (curbyt == 0xFF)
  {
   if ( (curbyt=getnext()) == 0xFD )
    outstr("bugl"); 
   else if (curbyt == 0xFE)
    outstr("bugw");
   else 
     { outstr(".byte   "); outnum((long)0xFF,2);
       outchar(','); outnum((long)curbyt,2); }
   goto end; 
  }
 else if ( !(cptr=vaxinst[curbyt]) )  
     { outstr(".byte   "); outnum((long)curbyt,2); goto end; }
    
 i=8;           /* Inialize blank count */
 while (*cptr && *cptr != ':')
  { outchar(*cptr++); i--; }
 if ( ! *cptr++ ) goto end;     /* Null-operand */

 while (i--) outchar(' ');

 while (i = *cptr++) { getarg(i); outchar(','); }

 --localoutbuf;

end: 
 outchar(0); return localpoint;
}

/*
 * largi - large ( >32 bit ) immediate delivery
 * It will work only for the hexadecimal notation so we can place
 * sequence of 32-bit numbers together.
 * It will work only if digits%8 == 0 !
 */

long largi1(digits)
int digits;
{
 long ltmp=getil();
 if ( (digits -= 8) && (largi1(digits) != 0) )
    leading++;          /* If already has a digits, switch off zero supression */
 outnum(ltmp,8);
 return ltmp;
}

largi(digits)
int digits;
{
 largi1(digits);
 leading = 0;   /* Reset default for output convertor */
}

/*
 * Kinds of operand:
 *
 * A - address
 * B - byte
 * D - floating point double precision
 * F - floating point single precision
 * G - floating point double precision, extended exponent
 * H - floating point quadruple precision
 * L - long word
 * O - octaword
 * Q - quadraword
 * W - word
 * V - bit field
 * b - byte branch offset
 * r - registers mask
 * w - word branch offset
 * x - CASEB \
 * y - CASEW  | instruction variable part (e.g. LIMIT followed by OFFSETs)
 * z - CASEL /
 */

static char *regnames[16] =
{
  "r0", "r1", "r2", "r3",
  "r4", "r5", "r6", "r7",
  "r8", "r9","r10","r11",
  "ap", "fp", "sp", "pc"
};

static outreg(regno)
int regno;
{ outstr(regnames[regno]); } 

getarg(kind)
char kind;
{
 long l,(*getfn)()=getil;
 int i,j,k,digcnt=8;
 char *str;

 isvalid=0;             /* No immediate value yet */
 switch(kind)
  {
   case 'x': getarg('B');
             goto casecom;

   case 'y': getarg('W');
             goto casecom;

   case 'z': getarg('L');
   casecom:  if (isvalid) isincase++; 
             svpoint = localpoint;
             return;

   case 'B': getfn=getib; digcnt=2; 
             break;

   case 'r': kind = 'W';
   case 'W': getfn=getiw; digcnt=4;
             break;

   case 'D':
   case 'G':
   case 'Q': digcnt=16;         /* Must be delivered as 2 longs */
             break;

   case 'H':
   case 'O': digcnt=32;         /* Must be delivered as 4 longs */
             break;

   case 'b': l=getib();
             outnum(localpoint+l,8);
             return;

   case 'w': l=getiw();
             outnum(localpoint+l,8);
             return;
  }

 k = (i=getnext()) & 0xF;       /* k is a register */
 switch (j = i >> 4)            /* j is a method */
  {

 case 4:                /* Indexed xxx[rx] */
   if (isindex)
    {
     outchar('?');
     outnum((long)i,2);
     outchar('?');
    }
   else
    {
     isindex++;
     getarg(kind);    
     outchar('['); outreg(k); outchar(']');
    }  
   isindex=0;
   break;

 case 5:                /* Register rx */
   outreg(k);
   break;

 case 7:                /* Register autodecrement -(rx) */
   outchar('-');
 case 6:                /* Register deferred (rx) */
 indreg:
   outchar('(');
   outreg(k);
   outchar(')');
   break;

 case 9:                /* Register deferred autoinc: @(rx)+ */
   outchar('@');
   getfn=getil; digcnt=8;
 case 8:                /* Register autoincrement (rx)+ */
   if (k==15)
    {
     if (j==8) outstr("I^");
     outchar('#');
     if (digcnt <= 8)
      {
       outnum((lastval=(*getfn)()),digcnt);
       isvalid++;
      }
     else largi(digcnt); /* Large immediate delivery */
    }
   else 
    {
     outchar('(');
     outreg(k);
     outstr(")+");
    }
   break;

 case 0xB:              /* @B^xx(rx) */
 case 0xA:              /*  B^xx(rx) */
   str="B^";
   getfn=getib; digcnt=2;
   goto indcom;

 case 0xD:              /* @W^yyyy(rx) */
 case 0xC:              /*  W^yyyy(rx) */
   str="W^";
   getfn=getiw; digcnt=4;
   goto indcom;

 case 0xF:              /* @^Lzzzzzzzz(rx) */
 case 0xE:              /*  ^Lzzzzzzzz(rx) */
   str="L^";
   getfn=getil; digcnt=8;
indcom:
   if (j&1) outchar('@');
   outstr(str);
   l=(*getfn)();
   if (k==15)
     outnum(localpoint+l,8);
   else
    {
     outnum(l,digcnt); 
     goto indreg;     
    }
   break;

default:        /* 00-3F literal :-) */
   outstr("S^#");
   outnum((lastval=(long)i),2);
   isvalid++; 
   break;

  }
}
